/* This is an implementation of shift scheduling problem
 * by S. Moriguchi and N. Tsuchimura.
 * This program is a part of demonstration CGI.
 * It is called from 'demo-shift.cgi' via 'odicon.cgi'.
 *
 *  c.f. G. Koole and E. van der Sluis (2003): Optimal shift scheduling
 *  with a global service level constraint, IIE Transactions, Vol. 35,
 *  pp. 1049-1055.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "odicon.h"


#define INTERVAL 13

/* determined in demo_shift_schedule() */
static double c = 0.01, mu = 20.0;
static double lambda[INTERVAL] = {88, 87, 86, 85, 55, 1, 1, 1, 1, 1, 1, 1, 1};
static int n_total;

/* x^n / n! */
static double exp_per_fact(double x, int n) {
    int i;
    double r = 1.0;

    for (i=n; i>=1; i--) r = r * x / i;
    return r;
}

/* pi = b / (a + b), which is safe even if b=inf */
static double pi(double a, double b) {
    if (b < a) {
        double s = b / a;
	return s / ( 1 + s );
    } else {
        double t = a / b;
	return 1 / ( t + 1 );
    }
}


static double gi(int n, int i) {
    double rho, rho_n, alpha, beta, gi;
    int j;
    
    rho = lambda[i-1] / mu;
    rho_n = rho / n;
 
    if (rho_n >=1) return PLUS_INFINITY;
    beta = exp_per_fact(rho,n) / (1-rho_n);
    
    alpha = 0;
    for (j=0; j<n; j++) {
        alpha += exp_per_fact(rho,j);
    }

    gi = lambda[i-1] * pi(alpha, beta) * exp(-n * mu*(1-rho_n)*c);

    /*
    printf(" gi[%d] = %g\trho=%g\ta=%g\tb=%g\tpi=%g\texp(*)=%g\n",
	   i, gi, rho, alpha, beta, pi(alpha, beta), exp(-n * mu*(1-rho_n)*c));
    */
    return gi;
}


/* f1(x) = \sum g_i */
static double f1(int dim, int x[]) {
    double r, LAMBDA;
    int j;
  
    assert(dim == 3);  /* check dimension */

    if (x[0] < 0) return PLUS_INFINITY;
    for (j=1; j<dim; j++) {
        if (x[j] < 0 || x[j]<x[j-1]) return PLUS_INFINITY;
    }
    for (j=0; j<dim; j++) {
        assert(lambda[j] < n_total * mu);
    }

    r = 0;
    r +=  gi(x[0]-0, 1);
    r +=  gi(x[0]-0, 2);
    r +=  gi(x[1]-0, 3);
    r +=  gi(x[1]-0, 4);
    r +=  gi(x[1]-0, 5);
    r +=  gi(x[2]-x[0], 6);
    r +=  gi(x[2]-x[0], 7);
    r +=  gi(x[2]-x[1], 8);
    r +=  gi(n_total-x[1], 9);
    r +=  gi(n_total-x[1], 10);
    r +=  gi(n_total-x[2], 11);
    r +=  gi(n_total-x[2], 12);
    r +=  gi(n_total-x[2], 13);

    LAMBDA = 0;
    for (j=0; j<INTERVAL; j++) {
        LAMBDA += lambda[j];
    }
    /*printf(" LAMBDA = %g\n\n", LAMBDA);*/
    
    return r / LAMBDA;
}

#define DIM 3

void demo_shift_schedule(const int dim) {
    int y[DIM+1] = {1, 1, 1, 7};  /* initial solution, multimodular */
    int x[DIM+1];                 /* L# variable */
    double ans;
    int i, sum, last;

    assert( dim == DIM );
    /* printf("gi = %f\n", gi(5,5)); */

    fvector_scan(1, &c);
    fvector_scan(1, &mu);
    fvector_scan(INTERVAL, lambda);
    ivector_scan(DIM+1, y);

    sum = 0;
    for (i=0; i<DIM+1; i++) {
        sum += y[i];
        x[i] = sum;
    }
    n_total = x[DIM];
    ans = lgconv_minimize(DIM, f1, x);

    /*printf("S(x) = %f\n", -ans);*/
    printf("\nN = %d\n", n_total);
    last = 0;
    for (i=0; i<DIM+1; i++) {
        y[i] = x[i] - last;
        last = x[i];
    }
    printf("\n");
    for (i=0; i<DIM+1; i++) {
        printf("x[%d] = %d\n", i+1, x[i]);
    }
    printf("\n");
    for (i=0; i<DIM+1; i++) {
        printf("y[%d] = %d\n", i+1, y[i]);
    }
}
