#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"

/* L, Lg, M, Mg, and others */
double quadratic(const int dim, int *x, double **a) {
    double ret = 0;
    int i, j;

    for (i=0; i<dim; i++) {
        double tmp = 0;
        for (j=0; j<dim; j++) {
            tmp += a[i][j] * x[j];
        }
        ret += tmp * x[i];
    }
    return ret/2;
}

double quadratic2(const int dim, int *x, double **a, double *b) {
    double ret = 0;
    int i;

    for (i=0; i<dim; i++) ret += x[i] * b[i];
    return quadratic(dim, x, a) + ret;
}

double quadratic3(const int dim, int *x, double **a, double *b, double c) {
    return quadratic2(dim, x, a, b) + c;
}

/* easy book p.41 (4.21) */
int lgconv_validate_hesse(const int dim, double **a) {
    int i, j;
    double sum;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) < 0) return FALSE;
    }
    return TRUE;
}

/* easy book p.42 (4.23) */
int lconv_validate_hesse(const int dim, double **a) {
    int i, j;
    double sum;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) != 0) return FALSE;
    }
    return TRUE;
}

/* easy book p.65 (5.15)(5.16) */
int mgconv_validate_hesse(const int dim, double **a) {
    int i, j, k;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = a[i][j];
            if (fcmp(tmp,a[j][i]) != 0) return FALSE;
            if (fcmp(tmp,0) < 0)        return FALSE;
            for (k=0; k<dim; k++) {
                if (k != i && k != j &&
                    fcmp(tmp,a[i][k]) < 0 &&
                    fcmp(tmp,a[j][k]) < 0) return FALSE;
            }
        }
    }
    return TRUE;
}

/* easy book p.68 (5.27) */
int mconv_validate_hesse(const int dim, double **a) {
    int i, j, k, l;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
    for (k=0; k<dim; k++) {
        if (k==i || k==j) continue;
        for (l=0; l<=k; l++) {
            if (l==i || l==j) continue;
            if (fcmp(a[i][j]+a[k][l],a[i][k]+a[j][l]) < 0 &&
                fcmp(a[i][j]+a[k][l],a[i][l]+a[j][k]) < 0) return FALSE;
        }
    }                    
        }
    }                    
    return TRUE;
}

/* L, Lg */
/* easy book p.42 (4.26) */
void make_hesse_lg(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  oracle_call++;  x[i]++;
            tmp += f(dim,x);  oracle_call++;  x[j]++;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;
            tmp += f(dim,x);  oracle_call++;  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
    /* diagonal elements */
    for (i=0; i<dim; i++) {
        double tmp = 0;
        tmp += f(dim,x);  oracle_call++;  x[i]++;
        tmp -= f(dim,x);  oracle_call++;  for (j=0; j<dim; j++) x[j]++;
        tmp += f(dim,x);  oracle_call++;  x[i]--;
        tmp -= f(dim,x);  oracle_call++;  for (j=0; j<dim; j++) x[j]--;
        for (j=0; j<dim; j++) if (i != j) tmp -= a[i][j];
        a[i][i] = tmp;
    }
}


/* M, Mg */
/* easy book p.69 (5.28) */
void make_hesse_mg(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  oracle_call++;  x[i]++;
            tmp += f(dim,x);  oracle_call++;  x[j]++;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;
            tmp += f(dim,x);  oracle_call++;  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
}


/* M */
/* Tree Metric */
void make_hesse_m(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            x[i]+=2;          tmp -= f(dim,x)/2;  oracle_call++;
            x[i]--;  x[j]++;  tmp += f(dim,x);    oracle_call++;
            x[i]--;  x[j]++;  tmp -= f(dim,x)/2;  oracle_call++;
                     x[j]-=2;
            a[i][j] = a[j][i] = tmp;
            /*
            tmp -= f(dim,x);  oracle_call++;  x[i]++;
            tmp += f(dim,x);  oracle_call++;  x[j]++;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;
            tmp += f(dim,x);  oracle_call++;  x[j]--;
            a[i][j] = a[j][i] = tmp;
            */
        }
    }
}


void make_liner_const(const int dim, double f(int,int*), int *x,
		      double **a, double *b, double *c) {
    int i, step = 10;
    double tmp;

    tmp = f(dim,x) - quadratic(dim,x,a);  oracle_call++;
    for (i=0; i<dim; i++) {
        x[i] += step;
        b[i] = (f(dim,x) - quadratic(dim,x,a) - tmp) / step;  oracle_call++;
        x[i] -= step;
    }

    *c = f(dim,x) - quadratic2(dim,x,a,b);  oracle_call++;
}


#ifdef TEST
int main(void) {
    double **a = fmatrix_new(2, 2);
    double  *b = fvector_new(2);
    int     *x = ivector_new(2);

    a[0][0] =  5; a[1][0] = -5;
    a[0][1] = -5; a[1][1] =  5;
    b[0] = b[1] = -1;

    x[0] = 1; x[1] = 1;

    printf("%g\n", quadratic2(2, x, a, b));

    free(x);
    free(b);
    free(a);
}
#endif
