/* sfm8.c --- IFF Scaling Algorithm for Submodular Function Minimization 

                     August 2000, December 2001, October 2003, 
                     May 2007, April 2009
                     
                     Coded by Satoru Iwata at University of Tokyo 
                                          and RIMS, Kyoto University  */ 

#include <stdio.h> 
#include <stdlib.h> 
#include <math.h>
#include <assert.h>
#include "sfm8.h"

#define EPS (1.0e-8)
#define K   (4*n)

/* compare double with error rounding */
static int fcmp( double a, double b ) {
    double diff;

    diff = a-b;
    if ( fabs(diff) < EPS ) return 0;
    return ( (diff>0) ? 1 : -1 );
}

static void *xmalloc( size_t size ) {
    void *s;

    if ( size == 0 ) size++;
    if ( (s=malloc(size)) == NULL ) {
        fprintf( stderr, "malloc : Not enough memory.\n" );
        exit( EXIT_FAILURE );
    }
    return s;
}

int* ivectoralloc(int n)   {  return xmalloc(sizeof(int) * n);  }

double* vectoralloc(int n) {  return xmalloc(sizeof(double) * n);  }

int** imatrixalloc(int m, int n) {
    int **a;
    int i, size1, size2, unit1, unit2;
    char *p;

    unit1 = sizeof(*a);
    unit2 = sizeof(**a);

    size1 = unit1 * m;
    size2 = unit2 * n;

    size1 = (size1+unit2-1) / unit2 * unit2; /* adjust alignment */

    p = xmalloc(size1 + size2 * m);

    a = (int**)p;  p += size1;
    for (i=0; i<m; i++) {
        a[i] = (int*)p;  p += size2;
    }
    return a;
}

double** matrixalloc(int m, int n) {
    double **a;
    int i, size1, size2, unit1, unit2;
    char *p;

    unit1 = sizeof(*a);
    unit2 = sizeof(**a);

    size1 = unit1 * m;
    size2 = unit2 * n;

    size1 = (size1+unit2-1) / unit2 * unit2; /* adjust alignment */

    p = xmalloc(size1 + size2 * m);

    a = (double**)p;  p += size1;
    for (i=0; i<m; i++) {
        a[i] = (double*)p;  p += size2;
    }
    return a;
}


int reduce(double **pt, double *lambda, int n, int k, int m, int *q){
   int    d,g,h,i,j,l,r,s;
   int    *p;
   double **a;
   double alpha, zeta, theta, gamma;
   double *kappa; 
   double rate;

   a=matrixalloc(m,n+1);
   p=ivectoralloc(n+1);
   kappa=vectoralloc(n);

   for (i=0;i<k;i++){
     for (j=0;j<n;j++){
       a[i][j]=pt[i][j];
     }
     a[i][n]=1.0;
   }   
   for (j=0;j<=n;j++){ 
     p[j]=j;
   }
   for (i=0,j=0;j<=n && i<k;i++){
     gamma=0.0;
     for (g=j,d=j;g<=n;g++){
       if (a[i][p[g]]>=0.0){
         alpha=a[i][p[g]];
       }
       else {
         alpha=-a[i][p[g]];
       }
       if (alpha>gamma){
         gamma=alpha;
         d=g;
       }
     }
     if (fcmp(gamma, 0.0)>0 && j<n){ 
       q[j]=i;
       s=p[d]; p[d]=p[j]; p[j]=s;
       for (g=0;g<j;g++){
         rate = a[i][p[g]] / a[i][p[j]];
         for (h=i+1;h<k;h++){
            a[h][p[g]] -= a[h][p[j]] * rate;
         }
         a[i][p[g]]=0.0;
       }

       for (g=n;g>j;g--){
         rate = a[i][p[g]] / a[i][p[j]];
         for (h=i+1;h<k;h++){
           a[h][p[g]] -= a[h][p[j]] * rate;
         }
         a[i][p[g]]=0.0;
       }
       j++;
     }
     else {
       theta=lambda[i];
       d=j; 
       for (h=0;h<j;h++){
         kappa[h]=a[i][p[h]]/a[q[h]][p[h]];
         if (fcmp(kappa[h],0.0)<0){
           zeta=-lambda[q[h]]/kappa[h];
           if (fcmp(zeta,theta)<0){
             theta=zeta;
             d=h;
	   }
         }
       }
       if (d<j){
         lambda[q[d]]=0.0;
         q[d]=i;
         kappa[d]=-1.0;
         for (g=0;g<d;g++){
           rate = a[i][p[g]] / a[i][p[d]];
           for (h=i+1;h<k;h++){
             a[h][p[g]] -= a[h][p[d]] * rate;
           }
         }
         for (g=n;g>d;g--){
             rate = a[i][p[g]] / a[i][p[d]];
             for (h=i+1;h<k;h++){
               a[h][p[g]] -= a[h][p[d]] * rate;
             }
           }
         }
       else {
         lambda[i]=0.0;
       }
       for (h=0;h<j;h++){
         lambda[q[h]]+=theta*kappa[h]; 
	 if (lambda[q[h]]<0) {
           lambda[q[h]]=0.0;
         }
       }
     }
   } 
   r=j;
   for (h=0;h<r;h++){
     while (q[h]<r && q[h]!=h){
	   d=q[h];
	   q[h]=q[d];
	   q[d]=d;
	 }
   }                        
   /*   gamma=0.0;
        for (l=0;l<k;l++){
          gamma+=lambda[l];
        }
        printf("sum of lambdas = %f", gamma);
        printf("       r=%d\n",r);           
   */
   free(kappa);
   free(p);
   free(a);
   return(r);
}


int sfm(int n, double oracle(int *p, int j, int n),
        double precision, int *mlst){ 
   int     **lst, **ord;
   double  **y;
   double  **f;
   double  **phi;
   int     *pa; 
   int     h,i,j,k,l,m,r,s,t,u,v;
   int     wr,ws,wt;
   int     *w;
   int     *q;
   double  *x;
   double  *z;
   double  *lambda;
   double  alpha, beta, gamma, delta, epsilon, eta, theta, zeta, mu, omega;
   int     count_oracle=0, count_reduce=0; 
   
   m=K+n; 
   lst=imatrixalloc(m,n);
   ord=imatrixalloc(m,n);
   y=matrixalloc(m,n);
   f=matrixalloc(m,n);
   phi=matrixalloc(n,n);
   pa=ivectoralloc(n+1);
   w=ivectoralloc(n);
   q=ivectoralloc(n);
   x=vectoralloc(n);
   z=vectoralloc(n);
   lambda=vectoralloc(m);

   for (s=0;s<n;s++){
     lst[0][s]=s;
   }
   omega=0.0;
   for (j=0;j<n;j++){
     count_oracle++;
     eta=oracle(lst[0],j+1,n);
     l=j;
     r=lst[0][j];
     s=r;
     for (h=j+1;h<n;h++){
       count_oracle++;
       lst[0][j]=lst[0][h];
       gamma=oracle(lst[0],j+1,n);
       /* printf("%f\n",gamma); */
       if (gamma<eta){
         eta=gamma;
         l=h;
         s=lst[0][h];
       }     
     }          
     lst[0][l]=r;
     lst[0][j]=s;
     ord[0][s]=j;
     f[0][j]=eta;
     if (eta<omega){
       omega=eta;
     }
   }
   s=lst[0][0];
   z[s]=f[0][0];
   x[s]=z[s]; 
   for (j=1;j<n;j++){
     s=lst[0][j];
     z[s]=f[0][j]-f[0][j-1];
     x[s]=z[s];
   }
   theta=0.0;
   zeta=0.0;
   for (s=0;s<n;s++){
     y[0][s]=z[s];
     if (z[s]<0){
       theta+=z[s];
     }
     else {
       zeta+=z[s];
     }
   }
   lambda[0]=1.0;
   if (-theta<zeta){
     zeta=-theta;
     for (u=0;u<n;u++){
       pa[u]=n;
     }
   }
   else {
     for (u=0;u<n;u++){
       pa[u]=-1;
     }
   }
   k=1;
   mu=n*n;
   epsilon=precision/mu;
   delta=zeta; 
   for (u=0;u<n;u++){
     for (v=0;v<n;v++){
       phi[u][v]=0.0;
     }
   }
   while (delta>epsilon){ 
     delta/=2;
     /* printf("delta %f\n",delta); */    
     for (u=0;u<n;u++){
       for (v=0;v<n;v++){
         if (phi[u][v]>delta){
           zeta=phi[u][v]-delta;
           phi[u][v]=delta;
           phi[v][u]=-delta;
           if (fcmp(phi[u][v],delta) > 0){
             printf("error: phi[%d][%d]=%f>delta=%f\n", u,v,phi[u][v], delta);
             exit(1);
           }
           z[u]-=zeta;
           z[v]+=zeta;
         }
       }
     }  
     while (1){
       for (u=0,wt=0;u<n;u++){
         if (z[u]<-delta){
           pa[u]=-1;
           w[wt]=u;
           wt++;
         }
         else {
           pa[u]=n;
         }
       }
       pa[n]=-1;
       ws=0;
       wr=0;
       while (ws<wt && pa[n]<0){
         s=w[ws];
         for (t=0;t<n;t++){
           if (pa[t]==n && phi[s][t]<=0.0){
             pa[t]=s;
             if (z[t]>delta){
               pa[n]=t;
               break;
             }
             w[wt]=t;
             wt++;
           }
         }
         ws++;
       }
       if (pa[n]<0){
         while (wr<wt){
           u=w[wr];
           for (i=0;i<k;i++){
             j=ord[i][u];
             if (j>0){         
               v=lst[i][j-1];
               if (pa[v]==n){
                 wr=wt;
                 pa[n]=n;
                 break;
               }             
	     }
           }
         wr++;
         }
         if (pa[n]==n){ 
           lst[i][j-1]=u;
           ord[i][u]=j-1;
           if (j>=n) { printf("error: j(=%d) >= n(=%d)\n", j, n); exit(1); }
           count_oracle++;
           gamma=oracle(lst[i],j,n);
           /* printf("%f\n",gamma); */
           if (gamma<omega){
             omega=gamma;
           }    
           zeta=f[i][j]-gamma;
           beta=y[i][v]-zeta;
           alpha=beta*lambda[i];
           if (fcmp(alpha,delta)>0){
             alpha=delta;
             lambda[k]=lambda[i];
             lambda[i]=alpha/beta;
             lambda[k]-=lambda[i];
             for (r=0;r<n;r++){
               y[k][r]=y[i][r];
               f[k][r]=f[i][r];
               lst[k][r]=lst[i][r];
               ord[k][r]=ord[i][r];
             }
             lst[k][j-1]=v;
             ord[k][u]=j;
             k++;
             if (k>=m) { printf("error: k(=%d) >= m(=%d)\n", k, m); exit(1); }
           }
           lst[i][j]=v;
           ord[i][v]=j;
           y[i][v]=zeta;
           y[i][u]+=beta;
           f[i][j-1]=gamma;
           x[v]-=alpha;
           x[u]+=alpha;  
           phi[u][v]-=alpha;
           phi[v][u]+=alpha;
         }
         else {
           break;
         }
       }
       else {
         t=pa[n];
         z[t]-=delta;
         s=pa[t];
         while (s>=0){
           if (!(0<=s && s<=n)) { printf("error: s=%d, n=%d\n", s,n); exit(1);}
           if (phi[s][t]>delta) { printf("error: phi[s][t]>delta\n"); exit(1);}
           phi[s][t]+=delta;
           phi[t][s]-=delta;
           t=s;
           s=pa[t];
         }
         z[t]+=delta;
         if (k>K){
           count_reduce++;
	   k=reduce(y,lambda,n,k,m,q);
	   for (i=0;i<k;i++){
             if (q[i]!=i){
               lambda[i]=lambda[q[i]];
               for (s=0;s<n;s++){
                 y[i][s]=y[q[i]][s];
                 f[i][s]=f[q[i]][s];
                 lst[i][s]=lst[q[i]][s];
                 ord[i][s]=ord[q[i]][s];
               }
             }
           }
         }
       }
     }
     if (k>n){
       count_reduce++;
       k=reduce(y,lambda,n,k,m,q);
       for (i=0;i<k;i++){
         if (q[i]!=i){
           lambda[i]=lambda[q[i]];
           for (s=0;s<n;s++){
             y[i][s]=y[q[i]][s];
             f[i][s]=f[q[i]][s];
             lst[i][s]=lst[q[i]][s];
             ord[i][s]=ord[q[i]][s];
           }
         }
       }
     } 
     theta=0.0;
     for (s=0;s<n;s++){
       if (x[s]<0){
         theta+=x[s];
       }
     }
   }
   /*
   printf("********************\n");
   printf("n=%d\n", n);
   printf("x^-(V) %f\n",theta);
   printf("min f  %f\n", omega); 
   printf("delta  %f\n", delta);
   printf("********************\n");
   printf("oracle %d\n", count_oracle);
   printf("reduce %d\n", count_reduce);
   printf("********************\n");
   */
   for (u=0,j=0;u<n;u++){
     if (pa[u]<n) {
       mlst[j]=u;
       j++;
     }
   }
   free(z);
   free(x);
   free(q);
   free(w);
   free(pa);
   free(phi);
   free(f);
   free(y);
   free(ord);
   free(lst);
   assert(fcmp( fabs(oracle(mlst,j,n)-omega), precision ) < 0);
   if (fcmp( fabs(oracle(mlst,j,n)-omega), precision * EPS ) > 0) {
       /* printf("omega = %10f\n", omega); */
       /* printf("oracle = %10f\n", oracle(mlst,j,n)); */
       return -1;
   }
   return(j);
}

int iff_call(int n, double oracle(int *p, int j, int n), int *mlst) {
    double precision = 1.0;
    int ret, i;

    for (i=0; i<10; i++) {
        ret = sfm(n, oracle, precision, mlst);
        if (ret >= 0) return ret;
        /* printf("precision (=%g) is too large.\n", precision); */
        precision /= 16;
    }
    printf("Small precision is not enough.\n");
    exit(1);
}


#ifdef TEST

double test(int *pt, int m, int n){
   double c,f;
   int j;

   c=0.0;
   for (j=0;j<m;j++){
     c+=5*pt[j]-2*n;
   }
   f=(n-m)*m-c;
   return(f);
} 

int main(){
   int m,n;
   double g;
   int *mlst;
 
   n=100;
   mlst=ivectoralloc(n);
   m=sfm(n,test,1.0,mlst);
   g=test(mlst,m,n);
   printf("min f %f\n", g);
   free(mlst);
   return(0);
}

#endif /* TEST */
